from __future__ import annotations
from typing import Callable
from AsymmetricEncryptions.General import XOR
from AsymmetricEncryptions.Protocols import KDF, PKCS7

class Feistel:

    def __init__(self, encryption_function: Callable[[bytes, bytes], bytes], block_size: int = 64, key_size: int = 32):
        """
        A basic Feistel network.
        @param key_size: the key size.
        @param encryption_function: The symmetric encryption function.
        @param block_size: The function's block size.
        """
        self.key_size = key_size
        assert block_size <= 256
        self.encryption_function = encryption_function
        self.block_size = block_size

    def expand_key(self, key: bytes, rounds: int) -> list[bytes]:
        """
        Expands a key.
        @param key: The key.
        @param rounds: The size of the keys
        @return: list[bytes]
        """
        current: bytes = KDF.derive_key(key)[:self.key_size]
        l: list[bytes] = [current]
        for _ in range(rounds - 1):
            l.append(KDF.derive_key(current)[:self.key_size])
        return l


    def encrypt(self, msg: bytes, key: bytes, *, rounds: int = 16, keys: list[bytes] = None) -> bytes:
        """
        Encrypts a message.
        @param msg: the message.
        @param key: the secret key.
        @param rounds: how many rounds (r > 4)
        @param keys: the round keys (No need to touch)
        @return: bytes, cipher text.
        """
        msg = PKCS7(self.block_size).pad(msg)
        if not keys:
            keys: list[bytes] = self.expand_key(key, rounds)
        half_point: int = int(self.block_size // 2)
        left: bytes = msg[0:half_point]
        right: bytes = msg[half_point:]
        for i in range(rounds):
            left = XOR.repeated_key_xor(left, self.encryption_function(right, keys[i % len(keys)]))
            tmp: bytes = left
            left = right
            right = tmp
        return right + left

    def decrypt(self, cipher: bytes, key: bytes, *, rounds: int = 16) -> bytes:
        """
        Decrypts a message.
        @param cipher: the cipher text.
        @param key: the private key.
        @param rounds: how many rounds the cipher text was encrypted with.
        @return: bytes, the message
        """
        keys: list[bytes] = self.expand_key(key, rounds)[::-1]
        return PKCS7(self.block_size).unpad(self.encrypt(cipher, key, rounds=rounds, keys=keys))



from __future__ import annotations
from .LWEKey import LWEKey
from AsymmetricEncryptions.General import BytesAndInts
import secrets


class LWE:
    """
    Learning With Errors
    """
    def __init__(self, key: LWEKey):
        self.key: LWEKey = key

    @staticmethod
    def encrypt_one_bit(key: LWEKey, m: int) -> tuple[list[int], int]:
        """
        Encrypts one bit, m will be either 0 or 1
        """
        m: int = m % 2
        number_of_samples: int = secrets.randbelow(key.q % len(key.A) + 1)
        while number_of_samples == 1: number_of_samples: int = secrets.randbelow(key.q % len(key.A) + 1)
        u: list[int] = []
        v: list[int] = []
        for _ in range(number_of_samples):
            r: int = secrets.randbelow(len(key.A))
            u.append(key.A[r])
            v.append(key.B[r])
        half_q_times_m: int = (key.q // 2) * m

        def sum_u(l) -> list[int]:
            result_arr: list[int] = [0] * len(l[0])
            for i in range(len(l)):
                for c in range(len(result_arr)):
                    result_arr[c] += l[i][c]
            return result_arr
        return sum_u(u), (sum(v) - half_q_times_m) % key.q

    def decrypt_one_bit(self, ciphertext: tuple[int, int]) -> int:
        """
        Decrypts a bit
        """
        def helper(A_i: list[int]) -> int:
            return sum([(A_i[i] * self.key.s[i]) % self.key.q for i in range(len(self.key.s))])
        u, v = ciphertext
        dec: int = (v - (helper(u))) % self.key.q
        return int(dec > (self.key.q // 2))


    @staticmethod
    def encrypt_message(key: LWEKey, message: bytes) -> list[tuple[int, int]]:
        """
        Encrypts a message (bytes)
        """
        m: int = BytesAndInts.byte2Int(message)
        out: list[int] = []
        for bit in format(m, "b"):
            bit = int(bit)
            out.append(LWE.encrypt_one_bit(key, bit))
        return out

    def decrypt_message(self, ciphertexts: list[tuple[int, int]]) -> bytes:
        """
        Decrypts a ciphertext
        """
        st: str = ""
        for index, i in enumerate(ciphertexts):
            st += str(self.decrypt_one_bit(i))
        return BytesAndInts.int2Byte(int(st, 2))



from __future__ import annotations
from . import ECPoint
import json

class ECCurve:

    def g(self) -> ECPoint.ECPoint:
        return ECPoint.ECPoint(curve=self, x=self.G.x, y=self.G.y)

    def set_g(self, G: ECPoint.ECPoint) -> None:
        self.G = G

    def infinity(self) -> ECPoint.ECPoint:
        return ECPoint.ECPoint(curve=self, x=None, y=None)

    def __init__(self, a: int, b: int, n: int, p: int, g: ECPoint.ECPoint = None):
        self.p = p
        self.n = n
        self.a = a
        self.b = b
        self.G = g

    def __eq__(self, other) -> bool:
        if not isinstance(other, ECCurve): return False
        return self.p == other.p and self.n == other.n and self.a == other.a and self.b == other.b and self.G == other.G

    def f(self, x: int):
        return (pow(x, 3, self.p) + self.a * x + self.b) % self.p

    def export(self) -> str:
        d: dict = {"a": self.a, "b": self.b, "n": self.n, "p": self.p, "g_x": self.G.x, "g_y": self.G.y}
        return json.dumps(d)

    @staticmethod
    def load(s: str):
        d: dict = json.loads(s)
        x, y = d["g_x"], d["g_y"]
        del d["g_x"]
        del d["g_y"]
        c: ECCurve = ECCurve(**d)
        c.set_g(ECPoint.ECPoint(c, x, y))
        return c



from AsymmetricEncryptions.Protocols.Feistel import Feistel
from hashlib import sha256


class FeistelSha256:

    @staticmethod
    def fSHA(msg, key):
        return sha256(msg + key).digest()

    @staticmethod
    def get_feistel() -> Feistel:
        return Feistel(FeistelSha256.fSHA)



__all__ = ["PublicPrivateKey", "Protocols", "Exceptions", "Unhazardous", "General"]
from .PublicPrivateKey import *
from .Protocols import *
from .Unhazardous import *
from .General import *
from .Exceptions import *


from __future__ import annotations
import secrets
from AsymmetricEncryptions.General import PrimeNumberGen, Exportation, XOR
import hashlib

class LWEKey:
    """
    Learning With Errors key object
    """

    def __init__(self, A: list[list[int]], B: list[int], q: int, e: list[int] = None, s: list[int] = None) -> None:
        self.A: list[list[int]] = A
        self.B: list[int] = B
        self.q: int = q
        self.e: list[int] = e
        self.s: list[int] = s
        self.is_private: bool = False
        self.public = None
        if s:
            self.is_private = True
            self.public = LWEKey(A, B, q)

    @staticmethod
    def new(nBit: int = 128, n: int = 1024, *, Svector_length: int = 32) -> LWEKey:
        """
        Generates a new Learning With Errors key
        """
        q: int = PrimeNumberGen.generate(nBit)
        s: list[int] = [secrets.randbelow(q) for _ in range(Svector_length)]
        A: list[int] = [[secrets.randbelow(q) for _ in range(Svector_length)] for _ in range(n)]
        error_margin: int = 4
        e: list[int] = [secrets.randbelow(error_margin) for _ in range(n)]

        def helper(A_i: list[int]) -> int:
            return sum([(A_i[i] * s[i]) % q for i in range(Svector_length)])
        B: list[int] = [(helper(A[i]) + e[i]) % q for i in range(n)]
        return LWEKey(A, B, q, e, s)

    def export(self, file_name: str, pwd: bytes = b"\x00", *, enc_func=XOR.repeated_key_xor) -> None:
        data_dict: dict = {"A": self.A, "B": self.B, "q": self.q, "e": self.e, "s": self.s}
        Exportation.export(file_name=file_name, pwd=pwd, data_dict=data_dict, exportation_func=enc_func)

    @staticmethod
    def load(file_name: str, pwd: bytes = b"\x00", *, dec_func=XOR.repeated_key_xor) -> LWEKey:
        return LWEKey(**Exportation.load(file_name=file_name, pwd=pwd, dec_func=dec_func))

    def __eq__(self, other: LWEKey) -> bool:
        if not isinstance(other, LWEKey): return False
        return hashlib.sha256(f"{self}".encode()).hexdigest() == hashlib.sha256(f"{other}".encode()).hexdigest()

    def __str__(self) -> str:
        r: str = ""
        if self.is_private:
            r += f"""
Private Key:

A = {self.A}
B = {self.B}
q = {self.q}
e = {self.e}
s = {self.s}


"""
        r += f"""
Public Key:

A = {self.A}
B = {self.B}
q = {self.q}

"""
        return r


__all__ = ["SSS",
           "OAEP",
           "FSZKP",
           "FiatShamirZeroKnowledgeProof",
           "OT1O2", "ObliviousTransfer",
           "Padding", "PKCS7", "RSADH", "DiffieHellman",
           "ThreePass", "ThreePassProtocol", "KDF", "SchnorrPOK", "POK", "YAK", "Feistel"]
from .SSS import SSS
from .OAEP import OAEP
from .FSZKP import FiatShamirZeroKnowledgeProof
from .OT1O2 import ObliviousTransfer
from .RSADH import DiffieHellman
from .KDF import KDF
from .ThreePass import ThreePassProtocol
from .Padding import PKCS7
from .SchnorrPOK import POK
from .YAK import YAK
from .Feistel import Feistel



class PKCS7(object):
    """PKCS 7 padding scheme object"""
    def __init__(self, block_size):
        self.block_size = block_size

    def pad(self, byte_str: bytes) -> bytes:
        """Pads a message"""
        padding_number: int = self.block_size - len(byte_str) % self.block_size
        if padding_number == self.block_size:
            return byte_str
        padding: bytes = chr(padding_number).encode() * padding_number
        return byte_str + padding

    def unpad(self, byte_str: bytes) -> bytes:
        if not byte_str: return byte_str
        if len(byte_str) % self.block_size:
            return byte_str
        padding_number: int = byte_str[-1]
        if padding_number >= self.block_size:
            return byte_str
        else:
            if all(padding_number == c for c in byte_str[-padding_number:]):
                return byte_str[0:-padding_number]
            else:
                return byte_str



from __future__ import annotations

from AsymmetricEncryptions.PublicPrivateKey.ElGamal import ElGamalKey, ElGamal
from AsymmetricEncryptions.Protocols import KDF, PKCS7
from AsymmetricEncryptions.General import BytesAndInts, XOR
import secrets
from hashlib import sha256

# A KEM for RSA

class UHElGamal:
    """
    A safe ElGamal encryption class
    """
    def __init__(self, private_key: ElGamalKey):
        """
        Private key for signing
        """
        self.private_key: ElGamalKey = private_key

    def encrypt(self, public_key: ElGamalKey, msg: bytes, *, encryption_function=XOR.repeated_key_xor, padding_block_size: int = 32) -> list[tuple[tuple[bytes], bytes], tuple[bytes, bytes, bytes]]:
        """
        Encrypts a message using ElGamal with a symmetric encryption method. (Uses the OTP without change, please use a better symmetric encryption algorithm such as AES-256)
        @rtype: The ciphertext as a tuple
        @param public_key: The public ElGamal key
        @param msg: The message you want to encrypt
        @param encryption_function: The symmetric encryption method, please also install PyCryptodome and use AES with here. The signature of the function must be (bytes, bytes) -> bytes
        @param padding_block_size: The padding block size, int
        """
        # pre reqs
        ciphertxt: list = []
        cipher: ElGamal = ElGamal(public_key)
        # Padding
        msg = PKCS7(padding_block_size).pad(msg)
        # Generates a random nonce
        r: bytes = BytesAndInts.int2Byte(secrets.randbelow(public_key.p-1))
        # encrypts r
        ciphertxt_r: bytes = cipher.encrypt(r)
        # Generates a key from the nonce
        key: bytes = KDF.derive_key(r)
        # symmetrically encrypts a message
        ciphertxt_msg: bytes = encryption_function(msg, key)
        # cipher text without a signature
        ciphertxt.append((ciphertxt_r, ciphertxt_msg))
        # Signer
        signer: ElGamal = ElGamal(self.private_key)
        # signing a hash
        signature = signer.sign(sha256(str(ciphertxt[0]).encode()).digest())
        ciphertxt.append(signature)
        return ciphertxt.copy()

    def decrypt(self, ciphertext: list[tuple[bytes, bytes], bytes], public_key: ElGamalKey, *, decryption_function=XOR.repeated_key_xor, padding_block_size: int = 32) -> bytes:
        """
        Decrypts a signed cipher text
        @rtype: bytes, message
        @param ciphertext: The cipher text.
        @param public_key: The sender's public key, for signature verification
        @param decryption_function: The symmetric decryption function
        @param padding_block_size: The padding block size, int
        """
        assert len(ciphertext) == 2
        # verifying the signature
        verifier: ElGamal = ElGamal(public_key)
        needed_signature: bytes = sha256(str(ciphertext[0]).encode()).digest()
        signature: tuple[bytes, bytes, bytes] = ciphertext[1]
        verifier.verify(signature, og_message=needed_signature)
        # decrypting
        cipher: ElGamal = ElGamal(self.private_key)
        ciphertext_r, ciphertext_msg = ciphertext[0]
        r: bytes = cipher.decrypt(ciphertext_r)
        key: bytes = KDF.derive_key(r)
        message: bytes = decryption_function(ciphertext_msg, key)
        message: bytes = PKCS7(padding_block_size).unpad(message)
        return message




from __future__ import annotations
from .ECCurve import ECCurve
from .ECPoint import ECPoint


class EllipticCurveNISTP256:
    """

    """
    p: int = 0xffffffff00000001000000000000000000000000ffffffffffffffffffffffff
    n: int = 0xffffffff00000000ffffffffffffffffbce6faada7179e84f3b9cac2fc632551
    a: int = 0xffffffff00000001000000000000000000000000fffffffffffffffffffffffc
    b: int = 0x5ac635d8aa3a93e7b3ebbd55769886bc651d06b0cc53b0f63bce3c3e27d2604b
    g_x: int = 0x6b17d1f2e12c4247f8bce6e563a440f277037d812deb33a0f4a13945d898c296
    g_y: int = 0x4fe342e2fe1a7f9b8ee7eb4a7c0f9e162bce33576b315ececbb6406837bf51f5

    @staticmethod
    def get_curve() -> ECCurve:
        c: ECCurve = ECCurve(EllipticCurveNISTP256.a, EllipticCurveNISTP256.b, EllipticCurveNISTP256.n, EllipticCurveNISTP256.p)
        G: ECPoint = ECPoint(c, EllipticCurveNISTP256.g_x, EllipticCurveNISTP256.g_y)
        c.set_g(G)
        return c


from __future__ import annotations
from . import ECPoint, ECKey
from AsymmetricEncryptions.General import XOR
from AsymmetricEncryptions.Protocols import KDF, PKCS7
import secrets

class ECIES:
    """Elliptic Curve Integrated Encryption Scheme.
    This is how encryption really works in ECC.
    """

    def __init__(self):
        pass



    @staticmethod
    def encrypt(msg: bytes, pub_key: ECPoint, encryption_function=XOR.repeated_key_xor, block_size: int=32) -> tuple[bytes, ECPoint]:
        """
        Encrypts a message with a public ECC key (As an ECPoint)
        @param msg: The message.
        @param pub_key: The public key as an ECPoint (If in ECKey you can just do .public_key)
        @param encryption_function: The symmetric encryption function used (XOR is used here if no changes are made)
        @return: The ciphertext as a tuple
        @param block_size: The symmetric block_size
        """
        msg: bytes = PKCS7(block_size).pad(msg)
        r: int = secrets.randbelow(pub_key.curve.p)
        G: ECPoint = pub_key.curve.g()
        R: ECPoint = G * r
        S: ECPoint = pub_key * r
        symmetric_key: bytes = bytes(S)
        symmetric_key: bytes = KDF.derive_key(symmetric_key)
        encrypted: bytes = encryption_function(msg, symmetric_key)
        return encrypted, R

    @staticmethod
    def decrypt(encrypted_msg, private_key: ECKey, decryption_function=XOR.repeated_key_xor, block_size=32) -> bytes:
        """
        Decrypts ciphertext.
        @param encrypted_msg: The ciphertext.
        @param private_key: The private key.
        @param decryption_function: The symmetric decryption function
        @return: message
        @param block_size: The symmetric decryption block size
        """
        encrypted, R = encrypted_msg
        S: ECPoint = R * private_key.private_key
        symmetric_key: bytes = bytes(S)
        symmetric_key: bytes = KDF.derive_key(symmetric_key)
        msg: bytes = decryption_function(encrypted, symmetric_key)
        msg: bytes = PKCS7(block_size).unpad(msg)
        return msg





class NeededValueIsNull(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)

class UnsafeEncryptionFunction(Exception):
    def __init__(self, func_name: str, msg: str):
        super().__init__(f"The function {func_name} is an unsafe function, error message:{msg}")


from __future__ import annotations
from hashlib import sha256
import secrets
from AsymmetricEncryptions.PublicPrivateKey.DLIES import DLIESKey
from AsymmetricEncryptions.General.BytesAndInts import BytesAndInts

# Schnorr Proof Of Knowledge (non-interactive)
class POK:

    def __init__(self, key: DLIESKey):
        self.key: DLIESKey = key

    def prove(self, msg: bytes) -> tuple[int, int, DLIESKey]:
        """
        Proves that you know a message without reveling what it is.
        @param msg: The message
        @return: Proof, tuple[int, int, DLIESKey]
        """
        new_key: DLIESKey = DLIESKey.build(self.key.g, self.key.n, BytesAndInts.byte2Int(msg))
        r: int = secrets.randbelow(new_key.n - 1)
        t: int = pow(new_key.g, r, new_key.n)
        c: int = POK.H([new_key.g, new_key.y, t])
        s: int = r + (new_key.x * c)
        return t, s, new_key.public

    @staticmethod
    def H(l: list[int]) -> int:
        nl: bytes = b"".join([BytesAndInts.int2Byte(i) for i in l])
        return BytesAndInts.byte2Int(sha256(nl).digest())

    @staticmethod
    def verify(proof: tuple[int, int, DLIESKey]) -> bool:
        """
        Verifies a proof
        @param proof: The proof
        @return: bool
        """
        t, s, key = proof
        c: int = POK.H([key.g, key.y, t])
        return pow(key.g, s, key.n) == (pow(key.y, c, key.n) * t) % key.n




from __future__ import annotations
from AsymmetricEncryptions.PublicPrivateKey.DLIES import DLIESKey
from AsymmetricEncryptions.Protocols import POK, KDF
from AsymmetricEncryptions.General import PrimeNumberGen, BytesAndInts
import secrets
import warnings
"""
The YAK is a public-key authenticated key-agreement protocol, proposed by Feng Hao in 2010.
It is claimed to be the simplest authenticated key exchange protocol among the related schemes, including MQV, HMQV, Station-to-Station protocol, SSL/TLS etc.
he authentication is based on public key pairs. As with other protocols, YAK normally requires a Public Key Infrastructure to distribute authentic public keys to the communicating parties. 
The security of YAK is disputed.
"""
class YAK:
    def __init__(self, g: int, q: int):
        self.g: int = g
        self.q: int = q

    @staticmethod
    def new(nBit: int = 2048) -> YAK:
        """
        Generates a new YAK object
        @param nBit: how big q is
        @return: YAK object
        """
        warnings.warn("YAK security is debatable")
        q: int = PrimeNumberGen.generate(nBit)
        g: int = secrets.randbelow(q)
        return YAK(g, q)

    def get_gq(self) -> tuple[int, int]:
        return self.g, self.q



    def stage_1(self) -> list[int, tuple[int, tuple[int, int, DLIESKey]]]:
        """
        First stage of the key exchange
        @return: list[int, tuple[int, tuple[int, int, DLIESKey]]]. [0] is private! Send only [1]!
        """
        x_or_y: int = secrets.randbelow(self.q - 1)
        g_x_or_y: int = self.helper(x_or_y)
        msg: bytes = BytesAndInts.int2Byte(x_or_y)
        dummy_key: DLIESKey = DLIESKey.build(self.g, self.q, x_or_y)
        proof: tuple[int, int, DLIESKey] = POK(dummy_key).prove(msg)
        return [x_or_y, (g_x_or_y, proof)]

    def helper(self, x: int) -> int:
        return pow(self.g, x, self.q)


    def stage_2(self, x_or_y: int, private_key: DLIESKey, other: tuple[int, tuple[int, int, DLIESKey]], public_key: DLIESKey) -> bytes:
        """
        The final stage of YAK
        @param x_or_y: The private int, (stage_1[0])
        @param private_key: You're private key
        @param other: stage_1[1]
        @param public_key: The other's public key.
        @return: Key as bytes
        """
        g_x_or_y, proof = other
        assert POK.verify(proof)
        K = pow((g_x_or_y * public_key.y), x_or_y + private_key.x, self.q)
        return KDF.derive_key(BytesAndInts.int2Byte(K))

